# tops.rb
# by pts@fazekas.hu at Tue Feb 13 16:23:08 CET 2001
# rewritten for singletons by pts@fazekas.hu at Wed Feb 14 19:19:51 CET 2001

$top=self

module TopUtil
  class TopFile
    # my policy is to add only methods which will be visible to the scripter.
    # methods for internal used should be put to `module TopUtil'
    attr_reader   :file  # String;
    attr_accessor :ltime # Time; # first load time (??) -- unused
    attr_accessor :xtime # Time, NilClass; # expiration time (when to be loaded again?)
    attr_accessor :mtime # Time;
    attr_reader   :deps  # Hash(String:TrueClass);
    def initialize(file_="unknown",now=Time.now)
      @file=file_
      @ltime=now
      @mtime=now
      @xtime=now
      @deps={}
    end
    def load_rb(fn, *args)
      TopUtil::load_rb fn, self, *args
    end
  end

  def dfsplit(df)
    # splits a dir/file to [dir, file]
    # Imp: ...
    raise RuntimeError, df unless ri=df.rindex("/")
    [df[0,ri],df[ri+1..-1]]
  end
  module_function :dfsplit
  def mk_absolute(tfn,s)
    # changes s in place
    # Imp: implement in PHP
    # raise TypeError unless tf.is_a? TopFile
    raise TypeError unless tfn.is_a? String
    raise TypeError unless s.is_a? String
    if (s[0]!=?/)
      dir=dfsplit(tfn)[0] # filename
      # PHP.print dir.inspect
      s.replace dir+"/"+s
    end
  end
  module_function :mk_absolute
  #def fn2mname(fn)
  #  "M_"+fn.gsub(/[^a-zA-Z0-9\/]/) { |s| sprintf "_%02X", s[0] }.gsub!(/\//, "__")
  #end
  #module_function :fn2mname
  
  @tfs={} # Hash(tfn:TopFile)

  def all_deps(tf)
    # TopFile tf; -- in @tfs
    # Module m;
    # Dat: not recursive!
    raise TypeError unless tf.is_a? TopFile
    h={}
    a=[tf.file]
    a.each { |fn|
      unless h.has_key? fn;
        h[fn]=true
        # Imp: check whether @tfs[fn] really exists...
        # Imp: re-think module removal
        a.concat @tfs[fn].deps.keys
      end
    }
    h.keys
  end
  module_function :all_deps
  def dirty?(tf,now)
    raise TypeError, tf.type.to_s unless tf.is_a? TopFile
    # Imp: check for file errors
    # return false if tf.file[-1]==?.
    return false unless tf.xtime
    # p tf.file, tf.mtime, File::mtime(tf.file)
    tf.xtime<=now and begin
      tf.mtime<File::mtime(tf.file)
    rescue SystemCallError
      true # dirty if not found!
    end
  end
  module_function :dirty?

  def load_rb(fn,tf=nil,*consts)
    # returns TopUtil::TopFile;
    raise TypeError unless not tf or tf.is_a? TopFile
    raise TypeError unless fn.is_a? String
    now=Time.now

    # PHP.print PHP.curfn.inspect
    mk_absolute(tf ? tf.file : PHP.curfn, fn)
    
    # fn.sub!(/(\.[^\.\/]*)?\z/,"") # strip extension
    # fn << ".rb"
    # Imp: more canonify
    
    if (newtf=@tfs[fn]) and !all_deps(newtf).find { |ifn| dirty?(@tfs[ifn],now) };
      # no need to reload. Keep original.
      STDERR.print "KEEP #{fn}.\n"
    else # (re)load from dist
      STDERR.print "RELOAD #{fn}.\n"
      # Imp: raise error if file not exist
      @tfs.delete fn
      # ^^^ Imp: possibly call callbacks if already existed
      @package=newtf=TopFile.new(fn,now)
      # Imp: set (delta) expiration time
      load fn, true # Ruby builtin, with wrapping
      # Imp: unset @package on LoadError
      # Imp: additional checks here before updating @tfs
      @tfs[fn]=newtf
      @package=nil
      # Imp: error if no `instance_eval' wrapper in fn
    end

    if tf;
      tf.deps[fn]=true
      consts.each { |c|
        raise TypeError unless c.is_a? String
        # Imp: allow Symbol
        # tf.const_set c, newtf.const_get c
        # PHP.print newtf.inspect
        $TMP=newtf.instance_eval c # Imp: brr
        tf.instance_eval "#{c} = \$TMP"
        # Imp: already init... warnings
        $TMP=nil
      }
    end
    newtf
  end
  def register(tf)
    # Imp: remove old registered...
    @tfs[tf.file]=tf
  end
  module_function :register
  
  def provide_fake_rb(fn)
    # String fn;
    # returns TopUtil::TopFile;
    # creates an empty TopFile which doesn't need to exist on disk.
    # fn must be an absolute (fake) filename!
    return @tfs[fn] if @tfs.has_key? fn
    @tfs[fn]=newtf=TopUtil::TopFile.new(fn)
    newtf.xtime=nil # never expires
    register newtf
    newtf
  end
  module_function :provide_fake_rb
  
  def sweep()
    # remove fake modules
    # Sun Feb 18 20:46:49 CET 2001
    # Imp: faster
    # Imp: only .php
    @tfs.delete_if { |k,v| not v.xtime }
    # ^^^ remove modules not on disk
  end
  module_function :sweep
  
  @package=nil
  def myself; @package; end
  module_function :mk_absolute
  module_function :load_rb
  module_function :myself
  module_function :dfsplit # for internal usage, too!
end

__END__

module PHP
  def print(*args); STDOUT.print(*args); end
  module_function :print
end

#root=TopUtil::TopFile.new(Dir.getwd+"/.root.")
#root.mtime=Time.now+10000 # Imp: farther in future...
#TopUtil::register root

root=TopUtil::fake_rb(Dir.getwd+"/root.php")

root.instance_eval do
  P1=TopUtil.load_rb(self,"p_1.rb","P1A","P2")
  # p P2A
  # p TopUtil::Mods::constants
  # p TopUtil::all_deps "#{Dirr}/#{File}"
  
  P22=TopUtil::load_rb(self,"p_2") # KEEP
  system "date; sleep 1; touch p_2.rb; echo TOUCHED"
  P22=TopUtil::load_rb(self,"p_2") # RELOAD
end

# p P22 # uninit...

# root.instance_eval do
#   P22=load_rb "p_2"
#   P22.korte
#   # p deps
#   load_rb "p_2"
#   #P22.korte
# end

# root.instance_eval do
#   p P22 # OK. But where is it stored...?
# end
# p root.const_get("P22") # nincs se const_get, se get_const
# p TopUtil::TopFile::P22 # uninitialized constant. !!
# p P22 # uninitialized constant. !!

__END__
